/*
  plugin.c: modules control
  Andriy Gritsenko (C) 2003
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

//#include "plugin.h"

#include "cfg.h"

typedef struct pl_record_s {
  struct pl_record_s *next;
  char *name;
  pl_info_t *info;
  struct pl_func_s *f;
  int type;
  cfg_tree_t *conf;
  int (*preopen)(pl_instance_t*,pl_instance_t*);
  //void *handle; // for loadable plugin
  int num_inst; // number of instances
} pl_record_t;

static pl_record_t *Plugins = NULL;

/* register module, returns 0 on success
   conf is config leaf of that module (should not be NULL!)
   preopen is module type specific function to fill instance priv structure
 */
int pl_register_plugin(pl_info_t *pl, cfg_tree_t *conf, int (*preopen)(pl_instance_t*,pl_instance_t*))
{
  pl_record_t *plr;

  if (!pl || !pl->name) return -1;
  //if (pl->version[0] != PLUGIN_VERSION_MAJOR || pl->version[1] < PLUGIN_VERSION_MINOR)
  //  return -1; // incorrect version
  for (plr = Plugins; plr; plr = plr->next) {
    if (pl->type == plr->type && !strcmp(pl->name, plr->name))
      break;
  }
  if (plr) return 0; // already registered
  plr = calloc(1, sizeof(pl_record_t));
  if (!plr) return -1; // out of memory
  plr->name = strdup(pl->name);
  plr->type = pl->type;
  plr->conf = conf;
  plr->preopen = preopen;
  //TODO: don't fill these if it's loadable plugin
  plr->info = pl;
  plr->f = pl->f;
  if (pl->opts)
    cfg_register(&conf, pl->opts, 0);
  //TODO: set num_inst to 1 if it's built-in plugin
  plr->next = Plugins;
  Plugins = plr;
  return 0;
}

/* get module instance by name and insert it in list,
   returns no-NULL on success */
pl_instance_t *pl_get_plugin(pl_instance_t *last, int type, char *name)
{
  pl_instance_t *pli = NULL;
  pl_record_t *pl;

  if (!name) return NULL;
  for (pl = Plugins; pl; pl = pl->next) {
    if (type == pl->type && !strcmp(name, pl->name))
      break;
  }
  if (!pl || !(pli = malloc(sizeof(pl_instance_t))))
    return NULL;
  //TODO: open dynamic loadable plugin and add opts if it's first instance
  //if (pl->info && pl->info->opts)
    //cfg_register(&pl->conf, pl->info->opts, 0);
  pl->num_inst++;
  pli->info = pl->info;
  pli->prev = last;
  pli->f = pl->f;
  pli->priv = NULL;
  if (pl->preopen && pl->preopen(pli, last) < 0) {
    free(pli);
    return NULL;
  }
  return pli;
}

/* open module, returns 0 on success */
int pl_open_plugin(pl_instance_t *pli)
{
  if (!pli || !pli->info || !pli->info->open)
    return -1; // panic - don't know how to open plugin
  return pli->info->open(pli);
}
